import logging
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import TYPE_CHECKING, Any
from uuid import uuid4

from griptape_nodes.retained_mode.managers.resource_components.comparator import Comparator

if TYPE_CHECKING:
    from griptape_nodes.retained_mode.managers.resource_components.resource_type import ResourceType

logger = logging.getLogger("griptape_nodes")

# Type aliases for resource requirements
RequirementValue = (
    Any  # Simple values: "present", 8, True
    | tuple[Any, str]  # Tuple format: (8, ">"), ("cuda", "present")
    | tuple[Any, Comparator]  # Enum format: (8, Comparator.GREATER_THAN)
)
Requirements = dict[str, RequirementValue]


class ResourceInstance(ABC):
    """Base class for resource instances that can be locked and managed."""

    def __init__(self, resource_type: "ResourceType", instance_id_prefix: str, capabilities: dict[str, Any]):
        self._resource_type = resource_type
        self._capabilities = deepcopy(capabilities)
        self._instance_id = f"{instance_id_prefix}_{uuid4()}"
        self._locked_by = None

    def get_resource_type(self) -> "ResourceType":
        """Get the resource type of this instance."""
        return self._resource_type

    def get_capabilities(self) -> list[str]:
        """Get the capability keys available for this resource."""
        return list(self._capabilities.keys())

    def has_capability(self, key: str) -> bool:
        """Check if resource has a specific capability key."""
        return key in self._capabilities

    def get_capability_value(self, key: str) -> Any:
        """Get a specific capability value.

        This method can be overridden by subclasses to support volatile resources
        that require real-time queries (e.g., current memory usage, GPU temperature)
        or resources that need external API calls to retrieve current values.

        The default implementation returns the static value from the capabilities dict.
        """
        return self._capabilities[key]

    def get_instance_id(self) -> str:
        """Unique handle/ID for this instance."""
        return self._instance_id

    def acquire_lock(self, owner_id: str) -> None:
        """Acquire exclusive lock.

        Args:
            owner_id: The ID of the entity requesting the lock

        Raises:
            ValueError: If the resource is already locked by another owner
        """
        if self._locked_by is not None:
            if self._locked_by == owner_id:
                # Already locked by the same owner - this is fine
                return
            msg = f"Resource {self._instance_id} is already locked by {self._locked_by}, cannot lock for {owner_id}"
            raise ValueError(msg)

        self._locked_by = owner_id
        logger.debug("Resource %s locked by %s", self._instance_id, owner_id)

    def release_lock(self, owner_id: str) -> None:
        """Release lock.

        Args:
            owner_id: The ID of the entity releasing the lock

        Raises:
            ValueError: If the resource is not locked by the specified owner
        """
        if self._locked_by != owner_id:
            if self._locked_by is None:
                msg = f"Resource {self._instance_id} is not locked, cannot release for {owner_id}"
            else:
                msg = f"Resource {self._instance_id} is locked by {self._locked_by}, cannot release for {owner_id}"
            raise ValueError(msg)

        self._locked_by = None
        logger.debug("Resource %s unlocked by %s", self._instance_id, owner_id)

    def is_locked(self) -> bool:
        """Check if resource is currently locked."""
        return self._locked_by is not None

    def is_locked_by(self, owner_id: str) -> bool:
        """Check if resource is locked by specific owner."""
        return self._locked_by == owner_id

    def get_lock_owner(self) -> str | None:
        """Get the current lock owner, if any."""
        return self._locked_by

    def force_unlock(self) -> None:
        """Force unlock this resource instance (for administrative operations)."""
        self._locked_by = None

    def get_all_capabilities_and_current_values(self) -> dict[str, Any]:
        """Get the complete capabilities dictionary with current values.

        This method returns all capability keys and their current values, which may
        include real-time queries for volatile resources that override get_capability_value().
        Returns a deep copy to prevent external modification of nested mutable objects.
        """
        # First get current values (may trigger real-time queries for volatile resources)
        result = {key: self.get_capability_value(key) for key in self._capabilities}
        # Then deep copy to protect against modification of nested mutable objects
        return deepcopy(result)

    def is_compatible_with(self, requirements: Requirements | None) -> bool:
        """Check if this instance can satisfy the requirements."""
        if requirements is None:
            return True

        for key, req_spec in requirements.items():
            # Handle both tuple and non-tuple formats
            if isinstance(req_spec, tuple):
                required_value, comparator_str = req_spec
            else:
                required_value, comparator_str = req_spec, Comparator.EQUALS

            # Convert string to StrEnum
            comparator = Comparator(comparator_str)

            if not self._compare_values(key, required_value, comparator):
                return False
        return True

    def _compare_values(self, key: str, required: Any, comparator: Comparator) -> bool:
        """Compare actual capability value against required value using comparator."""
        match comparator:
            case Comparator.NOT_PRESENT:
                return key not in self._capabilities
            case _:
                # For all other comparators, key must exist
                if key not in self._capabilities:
                    return False
                actual = self._capabilities[key]
                return self._apply_comparator(actual, required, comparator, key)

    def _apply_comparator(self, actual: Any, required: Any, comparator: Comparator, key: str) -> bool:
        """Apply the comparator operation."""
        match comparator:
            case Comparator.EQUALS:
                return actual == required
            case Comparator.NOT_EQUALS:
                return actual != required
            case (
                Comparator.GREATER_THAN_OR_EQUAL
                | Comparator.GREATER_THAN
                | Comparator.LESS_THAN_OR_EQUAL
                | Comparator.LESS_THAN
            ):
                return self._apply_numeric_comparator(actual, required, comparator)
            case Comparator.STARTS_WITH | Comparator.INCLUDES:
                return self._apply_string_comparator(actual, required, comparator)
            case Comparator.HAS_ANY | Comparator.HAS_ALL:
                return self._apply_container_comparator(actual, required, comparator, key)
            case Comparator.CUSTOM:
                return self._resource_type.handle_custom_requirement(
                    _instance=self,
                    _key=key,
                    _requirement_value=required,
                    _actual_value=actual,
                    _capabilities=self._capabilities,
                )
            case _:
                msg = f"Unknown comparator: {comparator}"
                raise ValueError(msg)

    def _apply_numeric_comparator(self, actual: Any, required: Any, comparator: Comparator) -> bool:
        """Apply numeric comparison operators."""
        match comparator:
            case Comparator.GREATER_THAN_OR_EQUAL:
                return actual >= required
            case Comparator.GREATER_THAN:
                return actual > required
            case Comparator.LESS_THAN_OR_EQUAL:
                return actual <= required
            case Comparator.LESS_THAN:
                return actual < required
            case _:
                msg = f"Invalid numeric comparator: {comparator}"
                raise ValueError(msg)

    def _apply_string_comparator(self, actual: Any, required: Any, comparator: Comparator) -> bool:
        """Apply string comparison operators."""
        match comparator:
            case Comparator.STARTS_WITH:
                return str(actual).startswith(str(required))
            case Comparator.INCLUDES:
                return str(required) in str(actual)
            case _:
                msg = f"Invalid string comparator: {comparator}"
                raise ValueError(msg)

    def _apply_container_comparator(self, actual: Any, required: Any, comparator: Comparator, key: str) -> bool:
        """Apply container comparison operators."""
        if not hasattr(actual, "__iter__") or isinstance(actual, str):
            msg = f"{comparator} comparator requires capability '{key}' to be a container, got {type(actual)}"
            raise ValueError(msg)
        if not hasattr(required, "__iter__") or isinstance(required, str):
            msg = f"{comparator} comparator requires requirement value to be a container, got {type(required)}"
            raise ValueError(msg)

        match comparator:
            case Comparator.HAS_ANY:
                return any(item in actual for item in required)
            case Comparator.HAS_ALL:
                return all(item in actual for item in required)
            case _:
                msg = f"Invalid container comparator: {comparator}"
                raise ValueError(msg)

    @abstractmethod
    def can_be_freed(self) -> bool:
        """Check if this resource can be safely freed."""

    @abstractmethod
    def free(self) -> None:
        """Free this resource instance."""



from __future__ import annotations

import hashlib
import logging
import threading
import uuid
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING

from watchfiles import Change, PythonFilter, watch

from griptape_nodes.drivers.storage.griptape_cloud_storage_driver import GriptapeCloudStorageDriver
from griptape_nodes.retained_mode.events.app_events import AppInitializationComplete
from griptape_nodes.retained_mode.events.base_events import AppEvent, ResultDetails
from griptape_nodes.retained_mode.events.sync_events import (
    StartSyncAllCloudWorkflowsRequest,
    StartSyncAllCloudWorkflowsResultFailure,
    StartSyncAllCloudWorkflowsResultSuccess,
    SyncComplete,
)
from griptape_nodes.retained_mode.events.workflow_events import (
    RegisterWorkflowsFromConfigRequest,
    RegisterWorkflowsFromConfigResultSuccess,
)
from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes

if TYPE_CHECKING:
    from griptape_nodes.retained_mode.events.base_events import ResultPayload
    from griptape_nodes.retained_mode.managers.config_manager import ConfigManager
    from griptape_nodes.retained_mode.managers.event_manager import EventManager


logger = logging.getLogger("griptape_nodes")


class SyncManager:
    """Manager for syncing workflows with cloud storage."""

    def __init__(self, event_manager: EventManager, config_manager: ConfigManager) -> None:
        self._active_sync_tasks: dict[str, threading.Thread] = {}
        self._watch_task: threading.Thread | None = None
        self._watching_stopped = threading.Event()
        self._config_manager = config_manager
        self._event_manager = event_manager

        # Hash tracking to prevent sync loops
        self._file_hashes = defaultdict(str)
        self._hash_lock = threading.Lock()

        # Initialize sync directory
        self._sync_dir = self._config_manager.workspace_path / self._config_manager.get_config_value(
            "synced_workflows_directory"
        )
        self._sync_dir.mkdir(parents=True, exist_ok=True)

        event_manager.assign_manager_to_request_type(
            StartSyncAllCloudWorkflowsRequest,
            self.on_start_sync_all_cloud_workflows_request,
        )

        event_manager.add_listener_to_app_event(
            AppInitializationComplete,
            self.on_app_initialization_complete,
        )

    def _set_expected_hash(self, path: str | Path, content: bytes) -> None:
        """Set the expected SHA256 hash for a file we're about to write.

        This should be called before writing a file to establish what content
        we expect to see, allowing us to distinguish our own writes from
        external modifications.

        Args:
            path: Path to the file that will be written
            content: The exact bytes that will be written to the file
        """
        from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes

        os_manager = GriptapeNodes.OSManager()
        path_str = str(os_manager.resolve_path_safely(Path(path)))
        file_hash = hashlib.sha256(content).hexdigest()
        with self._hash_lock:
            self._file_hashes[path_str] = file_hash
            logger.debug("Set expected hash for %s: %s", path_str, file_hash[:8])

    def _is_expected_content(self, path: str) -> bool:
        """Check if file content matches what we wrote (indicating a self-triggered event).

        This prevents sync loops by identifying when a file system event was caused
        by our own write operation rather than an external modification.

        Args:
            path: Path to the file to check

        Returns:
            True if the file content matches our expected hash (self-triggered event),
            False if it doesn't match or no expected hash exists (external change)
        """
        path_str = str(Path(path).resolve())
        with self._hash_lock:
            expected_hash = self._file_hashes.get(path_str)

        if not expected_hash:
            # No expected hash means this wasn't a self-triggered event
            return False

        try:
            actual_hash = hashlib.sha256(Path(path).read_bytes()).hexdigest()

            if actual_hash == expected_hash:
                # Content matches - this was our own write, clean up and ignore
                with self._hash_lock:
                    self._file_hashes.pop(path_str, None)
                logger.debug("File content matches expected hash for %s", path_str)
                return True
            # Content doesn't match - this is an external modification
            logger.debug("File content does not match expected hash for %s", path_str)
        except Exception as e:
            logger.debug("Error checking file hash for %s: %s", path_str, str(e))

        return False

    def on_start_sync_all_cloud_workflows_request(self, _request: StartSyncAllCloudWorkflowsRequest) -> ResultPayload:
        """Start syncing all cloud workflows to local synced_workflows directory."""
        try:
            storage_driver = self._get_cloud_storage_driver()
            sync_dir = self._sync_dir

            # List all assets in the bucket to get count
            files = storage_driver.list_files()
            workflow_files = [file for file in files if file.endswith(".py")]

            if not workflow_files:
                return StartSyncAllCloudWorkflowsResultSuccess(
                    sync_directory=str(sync_dir),
                    total_workflows=0,
                    result_details=ResultDetails(
                        message="No workflow files found in cloud storage.", level=logging.INFO
                    ),
                )

            # Start background sync with unique ID
            sync_task_id = str(uuid.uuid4())
            sync_thread = threading.Thread(
                target=self._sync_workflows_background,
                args=(sync_task_id, workflow_files, storage_driver, sync_dir),
                name=f"SyncWorkflows-{sync_task_id}",
                daemon=True,
            )

            self._active_sync_tasks[sync_task_id] = sync_thread
            sync_thread.start()
        except Exception as e:
            details = f"Failed to start cloud workflow sync: {e!s}"
            logger.error(details)
            return StartSyncAllCloudWorkflowsResultFailure(result_details=details)
        else:
            details = f"Started background sync for {len(workflow_files)} workflow files"
            return StartSyncAllCloudWorkflowsResultSuccess(
                sync_directory=str(sync_dir), total_workflows=len(workflow_files), result_details=details
            )

    def on_app_initialization_complete(self, _payload: AppInitializationComplete) -> None:
        """Automatically start syncing cloud workflows when the app initializes."""
        try:
            # Check if cloud storage is configured before attempting sync
            self._get_cloud_storage_driver()

            # Check if file watching is enabled before starting it
            enable_file_watching = self._config_manager.get_config_value("enable_workspace_file_watching", default=True)
            if enable_file_watching:
                # Start file watching after successful sync
                self._start_file_watching()
                logger.debug("File watching enabled - started watching synced workflows directory")
            else:
                logger.debug("File watching disabled - skipping file watching startup")

            logger.info("App initialization complete - starting automatic cloud workflow sync")

            # Create and handle the sync request
            sync_request = StartSyncAllCloudWorkflowsRequest()

            # Use handle_request to process through normal event system
            result = GriptapeNodes.handle_request(sync_request)

            if isinstance(result, StartSyncAllCloudWorkflowsResultSuccess):
                logger.info(
                    "Automatic cloud workflow sync started successfully - %d workflows will be synced to %s",
                    result.total_workflows,
                    result.sync_directory,
                )

            else:
                logger.debug("Automatic cloud workflow sync failed to start (likely cloud not configured)")

        except Exception as e:
            logger.debug("Automatic cloud workflow sync skipped: %s", str(e))

    def _get_cloud_storage_driver(self) -> GriptapeCloudStorageDriver:
        """Get configured cloud storage driver.

        Returns:
            Configured GriptapeCloudStorageDriver instance.

        Raises:
            RuntimeError: If required cloud configuration is missing.
        """
        secrets_manager = GriptapeNodes.SecretsManager()

        # Get cloud storage configuration from secrets
        bucket_id = secrets_manager.get_secret("GT_CLOUD_BUCKET_ID", should_error_on_not_found=False)
        base_url = secrets_manager.get_secret("GT_CLOUD_BASE_URL", should_error_on_not_found=False)
        api_key = secrets_manager.get_secret("GT_CLOUD_API_KEY")

        if not bucket_id:
            msg = "Cloud storage bucket_id not configured. Set GT_CLOUD_BUCKET_ID secret."
            raise RuntimeError(msg)
        if not api_key:
            msg = "Cloud storage api_key not configured. Set GT_CLOUD_API_KEY secret."
            raise RuntimeError(msg)

        workspace_directory = Path(self._config_manager.get_config_value("workspace_directory"))

        return GriptapeCloudStorageDriver(
            workspace_directory,
            bucket_id=bucket_id,
            base_url=base_url,
            api_key=api_key,
            static_files_directory=self._config_manager.get_config_value(
                "synced_workflows_directory", default="synced_workflows"
            ),
        )

    def _download_cloud_workflow_to_sync_dir(self, filename: str) -> bool:
        """Download a workflow file from cloud storage to the sync directory.

        Args:
            filename: Name of the workflow file to download from cloud

        Returns:
            True if download was successful, False otherwise
        """
        try:
            storage_driver = self._get_cloud_storage_driver()
            sync_dir = self._sync_dir

            # Download file content from cloud
            file_content = storage_driver.download_file(Path(filename))

            # Write to local sync directory
            local_file_path = sync_dir / filename

            # Check if file exists and has same content hash
            if local_file_path.exists():
                try:
                    existing_content = local_file_path.read_bytes()

                    existing_hash = hashlib.sha256(existing_content).hexdigest()
                    new_hash = hashlib.sha256(file_content).hexdigest()

                    if existing_hash == new_hash:
                        logger.debug("Skipping write - file already has same content hash: %s", filename)
                        return True
                except Exception as e:
                    logger.debug("Error checking existing file hash for %s: %s", filename, str(e))

            # Set expected hash before writing to prevent sync loops
            self._set_expected_hash(local_file_path, file_content)

            local_file_path.write_bytes(file_content)

            logger.info("Successfully downloaded cloud workflow to sync directory: %s", filename)
        except Exception as e:
            logger.error("Failed to download cloud workflow '%s': %s", filename, str(e))
            return False
        else:
            return True

    def _upload_workflow_file(self, file_path: Path) -> None:
        """Upload a single workflow file to cloud storage.

        Args:
            file_path: Path to the workflow file to upload.
        """
        try:
            # Check if valid workflow file
            if not file_path.name.endswith(".py"):
                logger.error("Invalid workflow file path: %s", file_path)
                return

            # Proceed with upload
            storage_driver = self._get_cloud_storage_driver()

            # Read file content
            file_content = file_path.read_bytes()

            # Upload to cloud storage using the upload_file method
            filename = file_path.name
            storage_driver.upload_file(Path(filename), file_content)

            logger.info("Successfully uploaded workflow file to cloud: %s", filename)

        except Exception as e:
            logger.error("Failed to upload workflow file '%s': %s", file_path.name, str(e))

    def _delete_workflow_file(self, file_path: Path) -> None:
        """Delete a workflow file from cloud storage.

        Args:
            file_path: Path to the workflow file that was deleted locally.
        """
        try:
            storage_driver = self._get_cloud_storage_driver()
            filename = file_path.name

            # Use the storage driver's delete method
            storage_driver.delete_file(Path(filename))
            logger.info("Successfully deleted workflow file from cloud: %s", filename)

        except Exception as e:
            logger.error("Failed to delete workflow file '%s' from cloud: %s", file_path.name, str(e))

    def _start_file_watching(self) -> None:
        """Start watching the synced_workflows directory for changes."""
        try:
            sync_dir = self._sync_dir

            # Stop any existing watching
            if self._watch_task and self._watch_task.is_alive():
                self._watching_stopped.set()
                self._watch_task.join(timeout=2.0)

            # Reset the stop event for new watching
            self._watching_stopped.clear()

            # Start new watching thread
            self._watch_task = threading.Thread(
                target=self._watch_files_thread,
                args=(str(sync_dir),),
                name="WatchFiles-SyncManager",
                daemon=True,
            )
            self._watch_task.start()

            logger.info("Started watching synced workflows directory: %s", sync_dir)

        except Exception as e:
            logger.error("Failed to start file watching: %s", str(e))

    def _watch_files_thread(self, sync_dir: str) -> None:
        """Background thread that watches for file changes using watchfiles."""
        try:
            logger.debug("File watching thread started for directory: %s", sync_dir)

            # Watch for changes in the sync directory using PythonFilter
            for changes in watch(sync_dir, watch_filter=PythonFilter(), stop_event=self._watching_stopped):
                if self._watching_stopped.is_set():
                    break

                for change, path_str in changes:
                    path = Path(path_str)

                    # Check if this was a self-triggered event for add/modify
                    if change in (Change.added, Change.modified) and self._is_expected_content(path_str):
                        logger.debug("Ignoring self-triggered %s event for: %s", change.name, path_str)
                        continue

                    # Handle the file change
                    self._handle_file_change(change, path, path_str)

        except Exception as e:
            if not self._watching_stopped.is_set():
                logger.error("Error in file watching thread: %s", str(e))
        finally:
            logger.debug("File watching thread stopped")

    def _handle_file_change(self, change: Change, path: Path, path_str: str) -> None:
        """Handle a file system change event."""
        if change == Change.added:
            logger.info("Detected external creation of workflow file: %s", path_str)
            self._upload_workflow_file(path)
        elif change == Change.modified:
            logger.info("Detected external modification of workflow file: %s", path_str)
            self._upload_workflow_file(path)
        elif change == Change.deleted:
            logger.info("Detected deletion of workflow file: %s", path_str)
            self._delete_workflow_file(path)

    def _download_single_workflow(
        self, file_name: str, storage_driver: GriptapeCloudStorageDriver, sync_dir: Path
    ) -> tuple[str, bool, str | None]:
        """Download a single workflow file.

        Args:
            file_name: Name of the workflow file to download from cloud
            storage_driver: Griptape Cloud storage driver instance for downloading
            sync_dir: Local directory path where the workflow file will be saved

        Returns:
            (filename, success, error_message)
        """
        try:
            # Download file content
            file_content = storage_driver.download_file(Path(file_name))

            # Extract just the filename (remove any directory prefixes)
            local_filename = Path(file_name).name
            local_file_path = sync_dir / local_filename

            # Check if file exists and has same content hash
            should_write = True
            if local_file_path.exists():
                try:
                    existing_content = local_file_path.read_bytes()

                    existing_hash = hashlib.sha256(existing_content).hexdigest()
                    new_hash = hashlib.sha256(file_content).hexdigest()

                    if existing_hash == new_hash:
                        logger.debug("Skipping write - file already has same content hash: %s", local_filename)
                        should_write = False
                except Exception as e:
                    logger.debug("Error checking existing file hash for %s: %s", local_filename, str(e))

            if should_write:
                # Set expected hash before writing to prevent sync loops
                self._set_expected_hash(local_file_path, file_content)

                # Write to local file
                local_file_path.write_bytes(file_content)

        except Exception as e:
            error_msg = str(e)
            logger.warning("Failed to sync workflow '%s': %s", file_name, error_msg)
            return file_name, False, error_msg
        else:
            logger.debug("Successfully synced workflow: %s", local_filename)
            return local_filename, True, None

    def _sync_workflows_background(
        self, sync_id: str, workflow_files: list[str], storage_driver: GriptapeCloudStorageDriver, sync_dir: Path
    ) -> None:
        """Background thread function to sync workflows."""
        synced_workflows = []
        failed_downloads = []
        total_workflows = len(workflow_files)

        logger.info("Starting background sync of %d workflows (sync_id: %s)", total_workflows, sync_id)

        # Use thread pool for concurrent downloads
        with ThreadPoolExecutor() as executor:
            # Submit all download tasks
            future_to_filename = {
                executor.submit(self._download_single_workflow, filename, storage_driver, sync_dir): filename
                for filename in workflow_files
            }

            # Collect results as they complete
            for future in future_to_filename:
                filename, success, _error = future.result()
                if success:
                    synced_workflows.append(filename)
                else:
                    failed_downloads.append(filename)

        if failed_downloads:
            logger.warning("Failed to sync %d workflows: %s", len(failed_downloads), failed_downloads)

        logger.info(
            "Background sync completed: %d of %d workflows synced to %s (sync_id: %s)",
            len(synced_workflows),
            len(workflow_files),
            sync_dir,
            sync_id,
        )

        # Emit sync complete event
        sync_complete_event = SyncComplete(
            sync_directory=str(sync_dir),
            synced_workflows=synced_workflows,
            failed_workflows=failed_downloads,
            total_workflows=total_workflows,
        )

        GriptapeNodes.EventManager().put_event(AppEvent(payload=sync_complete_event))

        # Register workflows from the synced directory
        if synced_workflows:
            logger.info("Registering %d synced workflows from configuration", len(synced_workflows))
            try:
                register_request = RegisterWorkflowsFromConfigRequest(
                    config_section="app_events.on_app_initialization_complete.workflows_to_register"
                )
                register_result = GriptapeNodes.handle_request(register_request)

                if isinstance(register_result, RegisterWorkflowsFromConfigResultSuccess):
                    logger.info(
                        "Successfully registered %d workflows after sync completion: %s",
                        len(register_result.succeeded_workflows),
                        register_result.succeeded_workflows,
                    )
                else:
                    logger.warning("Failed to register workflows after sync completion")
            except Exception as e:
                logger.error("Error registering workflows after sync: %s", str(e))

        # Clean up task tracking
        if sync_id in self._active_sync_tasks:
            del self._active_sync_tasks[sync_id]



from dataclasses import dataclass, field
from typing import Any

from griptape_nodes.retained_mode.events.base_events import (
    ExecutionPayload,
    RequestPayload,
    ResultDetails,
    ResultPayloadFailure,
    ResultPayloadSuccess,
    WorkflowAlteredMixin,
    WorkflowNotAlteredMixin,
)
from griptape_nodes.retained_mode.events.node_events import SerializedNodeCommands
from griptape_nodes.retained_mode.events.payload_registry import PayloadRegistry

# Requests and Results TO/FROM USER! These begin requests - and are not fully Execution Events.


@dataclass
@PayloadRegistry.register
class ResolveNodeRequest(RequestPayload):
    """Resolve (execute) a specific node.

    Use when: Running individual nodes, testing node execution, debugging workflows,
    stepping through execution manually. Validates inputs and runs node logic.

    Args:
        node_name: Name of the node to resolve/execute
        debug_mode: Whether to run in debug mode (default: False)

    Results: ResolveNodeResultSuccess | ResolveNodeResultFailure (with validation exceptions)
    """

    node_name: str
    debug_mode: bool = False


@dataclass
@PayloadRegistry.register
class ResolveNodeResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    """Node resolved successfully. Node execution completed and outputs are available."""


@dataclass
@PayloadRegistry.register
class ResolveNodeResultFailure(ResultPayloadFailure):
    """Node resolution failed. Contains validation errors that prevented execution.

    Args:
        validation_exceptions: List of validation errors that occurred
    """

    validation_exceptions: list[Exception]


@dataclass
@PayloadRegistry.register
class StartFlowRequest(RequestPayload):
    """Start executing a flow.

    Use when: Running workflows, beginning automated execution, testing complete flows.
    Validates all nodes and begins execution from resolved nodes.

    Args:
        flow_name: Name of the flow to start (deprecated, use flow_node_name)
        flow_node_name: Name of the flow node to start
        debug_mode: Whether to run in debug mode (default: False)

    Results: StartFlowResultSuccess | StartFlowResultFailure (with validation exceptions)
    """

    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None
    flow_node_name: str | None = None
    debug_mode: bool = False
    # If this is true, the final ControlFLowResolvedEvent will be pickled to be picked up from inside a subprocess.
    pickle_control_flow_result: bool = False


@dataclass
@PayloadRegistry.register
class StartFlowResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    """Flow started successfully. Execution is now running."""


@dataclass
@PayloadRegistry.register
class StartFlowResultFailure(ResultPayloadFailure):
    """Flow start failed. Contains validation errors that prevented execution.

    Args:
        validation_exceptions: List of validation errors that occurred
    """

    validation_exceptions: list[Exception]


@dataclass
@PayloadRegistry.register
class StartFlowFromNodeRequest(RequestPayload):
    """Start executing a flow from a specific node.

    Use when: Resuming execution from a particular node, debugging specific parts of a flow,
    re-running portions of a workflow, implementing custom execution control.

    Args:
        flow_name: Name of the flow to start (deprecated)
        node_name: Name of the node to start execution from
        debug_mode: Whether to run in debug mode (default: False)
        pickle_control_flow_result: If this is true, the final ControlFLowResolvedEvent will be pickled to be picked up from inside a subprocess

    Results: StartFlowFromNodeResultSuccess | StartFlowFromNodeResultFailure (with validation exceptions)
    """

    flow_name: str | None = None
    node_name: str | None = None
    debug_mode: bool = False
    pickle_control_flow_result: bool = False


@dataclass
@PayloadRegistry.register
class StartFlowFromNodeResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    """Flow started from node successfully. Execution is now running from the specified node."""


@dataclass
@PayloadRegistry.register
class StartFlowFromNodeResultFailure(ResultPayloadFailure):
    """Flow start from node failed. Contains validation errors that prevented execution.

    Args:
        validation_exceptions: List of validation errors that occurred
    """

    validation_exceptions: list[Exception]


@dataclass
@PayloadRegistry.register
class CancelFlowRequest(RequestPayload):
    """Cancel a running flow execution.

    Use when: Stopping long-running workflows, handling user cancellation,
    stopping execution due to errors or changes. Cleanly terminates execution.

    Args:
        flow_name: Name of the flow to cancel (deprecated)

    Results: CancelFlowResultSuccess | CancelFlowResultFailure (cancellation error)
    """

    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class CancelFlowResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    """Flow cancelled successfully. Execution has been terminated."""


@dataclass
@PayloadRegistry.register
class CancelFlowResultFailure(ResultPayloadFailure):
    """Flow cancellation failed. Common causes: flow not running, cancellation error."""


@dataclass
@PayloadRegistry.register
class UnresolveFlowRequest(RequestPayload):
    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class UnresolveFlowResultFailure(ResultPayloadFailure):
    pass


@dataclass
@PayloadRegistry.register
class UnresolveFlowResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    pass


# User Tick Events


# Step In: Execute one resolving step at a time (per parameter)
@dataclass
@PayloadRegistry.register
class SingleExecutionStepRequest(RequestPayload):
    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class SingleExecutionStepResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    pass


@PayloadRegistry.register
class SingleExecutionStepResultFailure(ResultPayloadFailure):
    pass


# Step Over: Execute one node at a time (execute whole node and move on) IS THIS CONTROL NODE OR ANY NODE?
@dataclass
@PayloadRegistry.register
class SingleNodeStepRequest(RequestPayload):
    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class SingleNodeStepResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    pass


@dataclass
@PayloadRegistry.register
class SingleNodeStepResultFailure(ResolveNodeResultFailure):
    pass


# Continue
@dataclass
@PayloadRegistry.register
class ContinueExecutionStepRequest(RequestPayload):
    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class ContinueExecutionStepResultSuccess(WorkflowAlteredMixin, ResultPayloadSuccess):
    pass


@dataclass
@PayloadRegistry.register
class ContinueExecutionStepResultFailure(ResultPayloadFailure):
    pass


@dataclass
@PayloadRegistry.register
class GetFlowStateRequest(RequestPayload):
    """Get the current execution state of a flow.

    Use when: Monitoring execution progress, debugging workflow state,
    implementing execution UIs, checking which nodes are active.

    Results: GetFlowStateResultSuccess (with control/resolving nodes) | GetFlowStateResultFailure (flow not found)
    """

    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class GetFlowStateResultSuccess(WorkflowNotAlteredMixin, ResultPayloadSuccess):
    """Flow execution state retrieved successfully.

    Args:
        control_nodes: Name of the current control node (if any)
        resolving_nodes: Name of the node currently being resolved (if any)
    """

    control_nodes: list[str]
    resolving_nodes: list[str]
    involved_nodes: list[str]


@dataclass
@PayloadRegistry.register
class GetFlowStateResultFailure(WorkflowNotAlteredMixin, ResultPayloadFailure):
    """Flow state retrieval failed. Common causes: flow not found, no current context."""


@dataclass
@PayloadRegistry.register
class GetIsFlowRunningRequest(RequestPayload):
    """Check if a flow is currently running.

    Use when: Monitoring execution status, preventing concurrent execution,
    implementing execution controls, checking if flow can be modified.

    Results: GetIsFlowRunningResultSuccess (with running status) | GetIsFlowRunningResultFailure (flow not found)
    """

    # Maintaining flow_name for backwards compatibility. Will be removed in https://github.com/griptape-ai/griptape-nodes/issues/1663
    flow_name: str | None = None


@dataclass
@PayloadRegistry.register
class GetIsFlowRunningResultSuccess(WorkflowNotAlteredMixin, ResultPayloadSuccess):
    """Flow running status retrieved successfully.

    Args:
        is_running: Whether the flow is currently executing
    """

    is_running: bool


@dataclass
@PayloadRegistry.register
class GetIsFlowRunningResultFailure(WorkflowNotAlteredMixin, ResultPayloadFailure):
    """Flow running status retrieval failed. Common causes: flow not found, no current context."""


# Execution Events! These are sent FROM the EE to the User/GUI. HOW MANY DO WE NEED?
@dataclass
@PayloadRegistry.register
class CurrentControlNodeEvent(ExecutionPayload):
    node_name: str


@dataclass
@PayloadRegistry.register
class CurrentDataNodeEvent(ExecutionPayload):
    node_name: str


@dataclass
@PayloadRegistry.register
class SelectedControlOutputEvent(ExecutionPayload):
    node_name: str
    selected_output_parameter_name: str


@dataclass
@PayloadRegistry.register
class ParameterSpotlightEvent(ExecutionPayload):
    node_name: str
    parameter_name: str


@dataclass
@PayloadRegistry.register
class ControlFlowResolvedEvent(ExecutionPayload):
    end_node_name: str
    parameter_output_values: dict
    # Optional field for pickled parameter values - when present, parameter_output_values contains UUID references
    unique_parameter_uuid_to_values: dict[SerializedNodeCommands.UniqueParameterValueUUID, Any] | None = field(
        default=None
    )


@dataclass
@PayloadRegistry.register
class ControlFlowCancelledEvent(ExecutionPayload):
    result_details: ResultDetails | str | None = None
    exception: Exception | None = None


@dataclass
@PayloadRegistry.register
class NodeResolvedEvent(ExecutionPayload):
    node_name: str
    parameter_output_values: dict
    node_type: str
    specific_library_name: str | None = None


@dataclass
@PayloadRegistry.register
class ParameterValueUpdateEvent(ExecutionPayload):
    node_name: str
    parameter_name: str
    data_type: str
    value: Any


@dataclass
@PayloadRegistry.register
class NodeUnresolvedEvent(ExecutionPayload):
    node_name: str


@dataclass
@PayloadRegistry.register
class NodeStartProcessEvent(ExecutionPayload):
    node_name: str


@dataclass
@PayloadRegistry.register
class NodeFinishProcessEvent(ExecutionPayload):
    node_name: str


@dataclass
@PayloadRegistry.register
class InvolvedNodesEvent(ExecutionPayload):
    """Event indicating which nodes are involved in the current execution.

    For parallel resolution: Dynamic list based on DAG builder state
    For control flow/sequential: All nodes when started, empty when complete
    """

    involved_nodes: list[str]


@dataclass
@PayloadRegistry.register
class GriptapeEvent(ExecutionPayload):
    node_name: str
    parameter_name: str
    type: str
    value: Any



"""Self command for Griptape Nodes CLI."""

import shutil
import sys

import typer

from griptape_nodes.cli.shared import (
    CONFIG_DIR,
    DATA_DIR,
    GITHUB_UPDATE_URL,
    LATEST_TAG,
    PYPI_UPDATE_URL,
    console,
)
from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
from griptape_nodes.utils.uv_utils import find_uv_bin
from griptape_nodes.utils.version_utils import (
    get_complete_version_string,
    get_current_version,
    get_latest_version_git,
    get_latest_version_pypi,
)

config_manager = GriptapeNodes.ConfigManager()
secrets_manager = GriptapeNodes.SecretsManager()
os_manager = GriptapeNodes.OSManager()

app = typer.Typer(help="Manage this CLI installation.")


@app.command()
def update() -> None:
    """Update the CLI."""
    _update_self()


@app.command()
def uninstall() -> None:
    """Uninstall the CLI."""
    _uninstall_self()


@app.command()
def version() -> None:
    """Print the CLI version."""
    _print_current_version()


def _get_latest_version(package: str, install_source: str) -> str:
    """Fetches the latest release tag from PyPI.

    Args:
        package: The name of the package to fetch the latest version for.
        install_source: The source from which the package is installed (e.g., "pypi", "git", "file").

    Returns:
        str: Latest release tag (e.g., "v0.31.4")
    """
    if install_source == "pypi":
        return get_latest_version_pypi(package, PYPI_UPDATE_URL)
    if install_source == "git":
        return get_latest_version_git(package, GITHUB_UPDATE_URL, LATEST_TAG)
    # If the package is installed from a file, just return the current version since the user is likely managing it manually
    return get_current_version()


def _update_self() -> None:
    """Installs the latest release of the CLI *and* refreshes bundled libraries."""
    console.print("[bold green]Starting updater...[/bold green]")

    os_manager.replace_process([sys.executable, "-m", "griptape_nodes.updater"])


def _print_current_version() -> None:
    """Prints the current version of the script."""
    version_string = get_complete_version_string()
    console.print(f"[bold green]{version_string}[/bold green]")


def _uninstall_self() -> None:
    """Uninstalls itself by removing config/data directories and the executable."""
    console.print("[bold]Uninstalling Griptape Nodes...[/bold]")

    # Remove config and data directories
    console.print("[bold]Removing config and data directories...[/bold]")
    dirs = [(CONFIG_DIR, "Config Dir"), (DATA_DIR, "Data Dir")]
    caveats = []
    for dir_path, dir_name in dirs:
        if dir_path.exists():
            console.print(f"[bold]Removing {dir_name} '{dir_path}'...[/bold]")
            try:
                shutil.rmtree(dir_path)
            except OSError as exc:
                console.print(f"[red]Error removing {dir_name} '{dir_path}': {exc}[/red]")
                caveats.append(
                    f"- [red]Error removing {dir_name} '{dir_path}'. You may want remove this directory manually.[/red]"
                )
        else:
            console.print(f"[yellow]{dir_name} '{dir_path}' does not exist; skipping.[/yellow]")

    # Handle any remaining config files not removed by design
    remaining_config_files = config_manager.config_files
    if remaining_config_files:
        caveats.append("- Some config files were intentionally not removed:")
        caveats.extend(f"\t[yellow]- {file}[/yellow]" for file in remaining_config_files)

    # If there were any caveats to the uninstallation process, print them
    if caveats:
        console.print("[bold]Caveats:[/bold]")
        for line in caveats:
            console.print(line)

    # Remove the executable
    console.print("[bold]Removing the executable...[/bold]")
    console.print("[bold yellow]When done, press Enter to exit.[/bold yellow]")

    # Remove the tool using UV
    uv_path = find_uv_bin()
    os_manager.replace_process([uv_path, "tool", "uninstall", "griptape-nodes"])



"""Models command for managing AI models."""

import asyncio
import sys
from typing import TYPE_CHECKING

import typer
from rich.table import Table

from griptape_nodes.cli.shared import console
from griptape_nodes.retained_mode.events.model_events import (
    DeleteModelDownloadRequest,
    DeleteModelDownloadResultFailure,
    DeleteModelDownloadResultSuccess,
    DeleteModelRequest,
    DeleteModelResultFailure,
    DeleteModelResultSuccess,
    ListModelDownloadsRequest,
    ListModelDownloadsResultFailure,
    ListModelDownloadsResultSuccess,
    ListModelsRequest,
    ListModelsResultFailure,
    ListModelsResultSuccess,
    ModelInfo,
    QueryInfo,
    SearchModelsRequest,
    SearchModelsResultFailure,
    SearchModelsResultSuccess,
)
from griptape_nodes.retained_mode.retained_mode import GriptapeNodes

if TYPE_CHECKING:
    from griptape_nodes.retained_mode.events.model_events import ModelDownloadStatus

app = typer.Typer(help="Manage AI models.")
downloads_app = typer.Typer(help="Manage model download tracking records.")

# Add downloads subcommand
app.add_typer(downloads_app, name="downloads")


@app.command("download")
def download_command(
    model_id: str = typer.Argument(..., help="Model ID or URL (e.g., 'microsoft/DialoGPT-medium')"),
    local_dir: str | None = typer.Option(None, "--local-dir", help="Local directory to download the model to"),
    revision: str = typer.Option("main", "--revision", help="Git revision to download"),
) -> None:
    """Download a model from Hugging Face Hub."""
    asyncio.run(_download_model(model_id, local_dir, revision))


@app.command("list")
def list_command() -> None:
    """List all downloaded model files in local cache."""
    asyncio.run(_list_models())


@app.command("delete")
def delete_command(
    model_id: str = typer.Argument(..., help="Model ID to delete (e.g., 'microsoft/DialoGPT-medium')"),
) -> None:
    """Delete model files from local cache."""
    asyncio.run(_delete_model(model_id))


@downloads_app.command("status")
def downloads_status_command(
    model_id: str = typer.Argument(None, help="Optional model ID to check download status for"),
) -> None:
    """Show download status for a specific model or all models."""
    asyncio.run(_get_model_status(model_id))


@downloads_app.command("list")
def downloads_list_command() -> None:
    """List all model download status records."""
    asyncio.run(_get_model_status(None))


@downloads_app.command("delete")
def downloads_delete_command(
    model_id: str = typer.Argument(
        ..., help="Model ID to delete download status for (e.g., 'microsoft/DialoGPT-medium')"
    ),
) -> None:
    """Delete download status tracking records for a model."""
    asyncio.run(_delete_model_status(model_id))


@app.command("search")
def search_command(
    query: str | None = typer.Argument(None, help="Search query to match against model names"),
    task: str | None = typer.Option(None, "--task", help="Filter by task type"),
    limit: int = typer.Option(20, "--limit", help="Maximum number of results (max: 100)"),
    sort: str = typer.Option("downloads", "--sort", help="Sort results by"),
    direction: str = typer.Option("desc", "--direction", help="Sort direction"),
) -> None:
    """Search for models on Hugging Face Hub."""
    asyncio.run(_search_models(query, task, limit, sort, direction))


async def _download_model(
    model_id: str,
    local_dir: str | None,
    revision: str,
) -> None:
    """Download a model from Hugging Face Hub.

    Args:
        model_id: Model ID or URL to download
        local_dir: Local directory to download the model to
        revision: Git revision to download
    """
    console.print(f"[bold green]Downloading model: {model_id}[/bold green]")

    try:
        # ModelManager DownloadModelRequest will use this command so it's important that we don't use the request ourselves
        model_manager = GriptapeNodes.ModelManager()
        local_path = model_manager.download_model(
            model_id=model_id,
            local_dir=local_dir,
            revision=revision,
            allow_patterns=None,
            ignore_patterns=None,
        )

        # Success case
        console.print("[bold green]Model downloaded successfully![/bold green]")
        console.print(f"[green]Downloaded to: {local_path}[/green]")

    except Exception as e:
        console.print("[bold red]Model download failed:[/bold red]")
        console.print(f"[red]{e}[/red]")
        sys.exit(1)


async def _list_models() -> None:
    """List all downloaded models in the local cache."""
    console.print("[bold green]Listing cached models...[/bold green]")

    # Create the list request
    request = ListModelsRequest()

    try:
        # Use the ModelManager to handle the listing
        result = await GriptapeNodes.ahandle_request(request)
        if isinstance(result, ListModelsResultSuccess):
            # Success case
            models = result.models
            if models:
                console.print(f"[bold green]Found {len(models)} cached models:[/bold green]")

                table = Table()
                table.add_column("Model ID", style="green")
                table.add_column("Size (GB)", style="cyan", justify="right")

                for model in models:
                    size_gb = round((model.size_bytes or 0) / (1024**3), 2) if model.size_bytes else 0.0
                    table.add_row(model.model_id, str(size_gb))
                console.print(table)
            else:
                console.print("[yellow]No models found in local cache[/yellow]")

        # Failure case

        elif isinstance(result, ListModelsResultFailure):
            console.print("[bold red]Model listing failed:[/bold red]")
            if result.result_details:
                console.print(f"[red]{result.result_details}[/red]")
            if result.exception:
                console.print(f"[dim]Error: {result.exception}[/dim]")
        else:
            console.print("[bold red]Model listing failed: Unknown error occurred[/bold red]")

    except Exception as e:
        console.print("[bold red]Unexpected error during model listing:[/bold red]")
        console.print(f"[red]{e}[/red]")


async def _delete_model(model_id: str) -> None:
    """Delete a model from the local cache.

    Args:
        model_id: Model ID to delete
    """
    console.print(f"[bold yellow]Deleting model: {model_id}[/bold yellow]")

    # Create the delete request
    request = DeleteModelRequest(model_id=model_id)

    try:
        # Use the ModelManager to handle the deletion
        result = await GriptapeNodes.ahandle_request(request)

        if isinstance(result, DeleteModelResultSuccess):
            # Success case
            console.print("[bold green]Model deleted successfully![/bold green]")
            console.print(f"[green]Deleted: {result.deleted_path}[/green]")
        # Failure case

        elif isinstance(result, DeleteModelResultFailure):
            console.print("[bold red]Model deletion failed:[/bold red]")
            if result.result_details:
                console.print(f"[red]{result.result_details}[/red]")
            if result.exception:
                console.print(f"[dim]Error: {result.exception}[/dim]")
        else:
            console.print("[bold red]Model deletion failed: Unknown error occurred[/bold red]")

    except Exception as e:
        console.print("[bold red]Unexpected error during model deletion:[/bold red]")
        console.print(f"[red]{e}[/red]")


def _format_download_row(download: "ModelDownloadStatus") -> tuple[str, str, str, str, str, str]:
    """Format a download status object into table row data.

    Args:
        download: ModelDownloadStatus object

    Returns:
        tuple: (model_id, status_colored, progress_str, size_str, eta_str, started_str)
    """
    progress_str = _format_progress(download)
    size_str = _format_size(download)
    eta_str = _format_eta(download)
    started_str = _format_timestamp(download)
    status_colored = _format_status(download)

    return (
        download.model_id,
        status_colored,
        progress_str,
        size_str,
        eta_str,
        started_str,
    )


def _format_progress(download: "ModelDownloadStatus") -> str:
    """Format download progress information."""
    if download.total_files is not None and download.completed_files is not None and download.total_files > 0:
        progress_percent = (download.completed_files / download.total_files) * 100
        return f"{progress_percent:.1f}%"
    return "Unknown"


def _format_size(download: "ModelDownloadStatus") -> str:
    """Format download size information."""
    if download.total_files is not None and download.completed_files is not None:
        return f"{download.completed_files}/{download.total_files} files"
    return "Unknown"


def _format_eta(download: "ModelDownloadStatus") -> str:
    """Format estimated time of arrival."""
    # ETA is not available in the current ModelDownloadStatus structure
    # For active downloads, we could potentially calculate based on progress
    # but without timing data, we return a status-appropriate message
    if download.status == "downloading":
        return "In progress"
    if download.status == "completed":
        return "Completed"
    if download.status == "failed":
        return "Failed"
    return "Unknown"


def _format_timestamp(download: "ModelDownloadStatus") -> str:
    """Format started timestamp."""
    started_at = download.started_at
    if not started_at:
        return "Unknown"

    try:
        from datetime import datetime

        dt = datetime.fromisoformat(started_at)
        return dt.strftime("%H:%M:%S")
    except Exception:
        return started_at[:10]  # Fallback


def _format_status(download: "ModelDownloadStatus") -> str:
    """Format status with color coding."""
    status = download.status
    status_colors = {
        "completed": "green",
        "failed": "red",
        "downloading": "yellow",
    }

    if status in status_colors:
        return f"[{status_colors[status]}]{status}[/{status_colors[status]}]"
    return status


def _display_downloads_table(downloads: list["ModelDownloadStatus"]) -> None:
    """Display downloads in a formatted table.

    Args:
        downloads: List of ModelDownloadStatus objects
    """
    console.print(f"[bold green]Found {len(downloads)} download(s):[/bold green]")

    table = Table()
    table.add_column("Model ID", style="green")
    table.add_column("Status", style="cyan")
    table.add_column("Progress", style="yellow", justify="right")
    table.add_column("Size", style="blue", justify="right")
    table.add_column("ETA", style="magenta", justify="right")
    table.add_column("Started", style="dim")

    for download in downloads:
        row_data = _format_download_row(download)
        table.add_row(*row_data)

    console.print(table)


async def _get_model_status(model_id: str | None) -> None:
    """Get download status for models.

    Args:
        model_id: Optional model ID to get status for
    """
    if model_id:
        console.print(f"[bold green]Getting download status for: {model_id}[/bold green]")
    else:
        console.print("[bold green]Getting download status for all models...[/bold green]")

    # Create the status request
    request = ListModelDownloadsRequest(model_id=model_id)

    try:
        # Use the ModelManager to handle the status query
        result = await GriptapeNodes.ahandle_request(request)

        if isinstance(result, ListModelDownloadsResultSuccess):
            # Success case
            downloads = result.downloads
            if downloads:
                _display_downloads_table(downloads)
            elif model_id:
                console.print(f"[yellow]No download found for model: {model_id}[/yellow]")
            else:
                console.print("[yellow]No downloads found[/yellow]")

        elif isinstance(result, ListModelDownloadsResultFailure):
            console.print("[bold red]Failed to get download status:[/bold red]")
            if result.result_details:
                console.print(f"[red]{result.result_details}[/red]")
            if result.exception:
                console.print(f"[dim]Error: {result.exception}[/dim]")
        else:
            console.print("[bold red]Failed to get download status: Unknown error occurred[/bold red]")

    except Exception as e:
        console.print("[bold red]Unexpected error getting download status:[/bold red]")
        console.print(f"[red]{e}[/red]")


async def _search_models(
    query: str | None,
    task: str | None,
    limit: int,
    sort: str,
    direction: str,
) -> None:
    """Search for models on Hugging Face Hub.

    Args:
        query: Search query to match against model names
        task: Filter by task type
        limit: Maximum number of results
        sort: Sort results by
        direction: Sort direction
    """
    if query:
        console.print(f"[bold green]Searching for models: {query}[/bold green]")
    else:
        console.print("[bold green]Searching for models...[/bold green]")

    # Create the search request
    request = SearchModelsRequest(
        query=query,
        task=task,
        library=None,
        author=None,
        tags=None,
        limit=limit,
        sort=sort,
        direction=direction,
    )

    try:
        # Use the ModelManager to handle the search
        result = await GriptapeNodes.ahandle_request(request)

        if isinstance(result, SearchModelsResultSuccess):
            # Success case
            models = result.models
            if models:
                _display_search_results(models, result.query_info)
            else:
                console.print("[yellow]No models found matching the search criteria[/yellow]")

        elif isinstance(result, SearchModelsResultFailure):
            console.print("[bold red]Model search failed:[/bold red]")
            if result.result_details:
                console.print(f"[red]{result.result_details}[/red]")
            if result.exception:
                console.print(f"[dim]Error: {result.exception}[/dim]")
        else:
            console.print("[bold red]Model search failed: Unknown error occurred[/bold red]")

    except Exception as e:
        console.print("[bold red]Unexpected error during model search:[/bold red]")
        console.print(f"[red]{e}[/red]")


def _display_search_results(models: list[ModelInfo], query_info: QueryInfo) -> None:
    """Display model search results in a formatted table.

    Args:
        models: List of model information
        query_info: Information about the search query
    """
    console.print(f"[bold green]Found {len(models)} models[/bold green]")

    # Show search parameters if any were used
    params = []
    if query_info.query:
        params.append(f"query: {query_info.query}")
    if query_info.task:
        params.append(f"task: {query_info.task}")
    if query_info.library:
        params.append(f"library: {query_info.library}")
    if query_info.author:
        params.append(f"author: {query_info.author}")
    if query_info.tags:
        params.append(f"tags: {', '.join(query_info.tags)}")

    if params:
        console.print(f"[dim]Search parameters: {', '.join(params)}[/dim]")

    table = Table()
    table.add_column("Model ID", style="green")
    table.add_column("Author", style="blue")
    table.add_column("Downloads", style="cyan", justify="right")
    table.add_column("Likes", style="yellow", justify="right")
    table.add_column("Task", style="magenta")
    table.add_column("Library", style="dim")

    for model in models:
        downloads_str = f"{model.downloads:,}" if model.downloads else "0"
        likes_str = str(model.likes or 0)
        task_str = model.task or ""
        library_str = model.library or ""
        author_str = model.author or ""

        table.add_row(
            model.model_id,
            author_str,
            downloads_str,
            likes_str,
            task_str,
            library_str,
        )

    console.print(table)


async def _delete_model_status(model_id: str) -> None:
    """Delete download status records for a model.

    Args:
        model_id: Model ID to delete download status for
    """
    console.print(f"[bold yellow]Deleting download status for: {model_id}[/bold yellow]")

    # Create the delete request
    request = DeleteModelDownloadRequest(model_id=model_id)

    try:
        # Use the ModelManager to handle the deletion
        result = await GriptapeNodes.ahandle_request(request)

        if isinstance(result, DeleteModelDownloadResultSuccess):
            # Success case
            console.print("[bold green]Download status deleted successfully![/bold green]")
            console.print(f"[green]Deleted status file: {result.deleted_path}[/green]")
        # Failure case

        elif isinstance(result, DeleteModelDownloadResultFailure):
            console.print("[bold red]Download status deletion failed:[/bold red]")
            if result.result_details:
                console.print(f"[red]{result.result_details}[/red]")
            if result.exception:
                console.print(f"[dim]Error: {result.exception}[/dim]")
        else:
            console.print("[bold red]Download status deletion failed: Unknown error occurred[/bold red]")

    except Exception as e:
        console.print("[bold red]Unexpected error during download status deletion:[/bold red]")
        console.print(f"[red]{e}[/red]")



"""Engine command for Griptape Nodes CLI."""

from rich.prompt import Confirm

from griptape_nodes.app import start_app
from griptape_nodes.cli.commands.init import _run_init
from griptape_nodes.cli.commands.self import _get_latest_version, _update_self
from griptape_nodes.cli.shared import (
    CONFIG_DIR,
    ENV_API_KEY,
    ENV_GTN_BUCKET_NAME,
    ENV_LIBRARIES_SYNC,
    ENV_REGISTER_ADVANCED_LIBRARY,
    ENV_STORAGE_BACKEND,
    ENV_WORKSPACE_DIRECTORY,
    PACKAGE_NAME,
    InitConfig,
    console,
)
from griptape_nodes.utils.version_utils import get_current_version, get_install_source


def engine_command() -> None:
    """Run the Griptape Nodes engine."""
    _start_engine()


def _start_engine() -> None:
    """Starts the Griptape Nodes engine."""
    if not CONFIG_DIR.exists():
        # Default init flow if there is no config directory
        console.print("[bold green]Config directory not found. Initializing...[/bold green]")
        _run_init(
            InitConfig(
                workspace_directory=ENV_WORKSPACE_DIRECTORY,
                api_key=ENV_API_KEY,
                storage_backend=ENV_STORAGE_BACKEND,
                register_advanced_library=ENV_REGISTER_ADVANCED_LIBRARY,
                interactive=True,
                config_values=None,
                secret_values=None,
                libraries_sync=ENV_LIBRARIES_SYNC,
                bucket_name=ENV_GTN_BUCKET_NAME,
            )
        )

    console.print("[bold green]Starting Griptape Nodes engine...[/bold green]")
    start_app()


def _auto_update_self() -> None:
    """Automatically updates the script to the latest version if the user confirms."""
    console.print("[bold green]Checking for updates...[/bold green]")
    source, commit_id = get_install_source()
    current_version = get_current_version()
    latest_version = _get_latest_version(PACKAGE_NAME, source)

    if source == "git" and commit_id is not None:
        can_update = commit_id != latest_version
        update_message = f"Your current engine version, {current_version} ({source} - {commit_id}), doesn't match the latest release, {latest_version}. Update now?"
    else:
        can_update = current_version < latest_version
        update_message = f"Your current engine version, {current_version}, is behind the latest release, {latest_version}. Update now?"

    if can_update:
        update = Confirm.ask(update_message, default=True)

        if update:
            _update_self()



"""Format specifier classes for macro variable transformations."""

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from dataclasses import dataclass

from griptape_nodes.common.macro_parser.exceptions import MacroResolutionError, MacroResolutionFailureReason


@dataclass
class FormatSpec(ABC):
    """Base class for format specifiers."""

    @abstractmethod
    def apply(self, value: str | int) -> str | int:
        """Apply this format spec to a value during resolution.

        Args:
            value: Value to transform

        Returns:
            Transformed value

        Raises:
            MacroResolutionError: If format cannot be applied to value type

        Examples:
            >>> # NumericPaddingFormat(width=3).apply(5)
            "005"
            >>> # LowerCaseFormat().apply("MyWorkflow")
            "myworkflow"
        """

    @abstractmethod
    def reverse(self, value: str) -> str | int:
        """Reverse this format spec during matching (best effort).

        Args:
            value: Formatted string value from a path

        Returns:
            Original value before format was applied

        Raises:
            MacroResolutionError: If value cannot be reversed

        Examples:
            >>> # NumericPaddingFormat(width=3).reverse("005")
            5
            >>> # SeparatorFormat(separator="_").reverse("workflow_")
            "workflow"
        """


@dataclass
class SeparatorFormat(FormatSpec):
    """Separator appended to variable value like :_, :/, :foo.

    Must be first format spec in list (if present).
    Syntax: {var:_} or {var:'lower'} (quotes to disambiguate from transformations)
    """

    separator: str  # e.g., "_", "/", "foo"

    def apply(self, value: str | int) -> str:
        """Append separator to value."""
        return str(value) + self.separator

    def reverse(self, value: str) -> str:
        """Remove separator from end of value."""
        if value.endswith(self.separator):
            return value[: -len(self.separator)]
        return value


@dataclass
class NumericPaddingFormat(FormatSpec):
    """Numeric padding format like :03, :04."""

    width: int  # e.g., 3 for :03

    def apply(self, value: str | int) -> str:
        """Apply numeric padding: 5 → "005"."""
        if not isinstance(value, int):
            if not str(value).isdigit():
                msg = (
                    f"Numeric padding format :{self.width:0{self.width}d} "
                    f"cannot be applied to non-numeric value: {value}"
                )
                raise MacroResolutionError(
                    msg,
                    failure_reason=MacroResolutionFailureReason.NUMERIC_PADDING_ON_NON_NUMERIC,
                )
            value = int(value)
        return f"{value:0{self.width}d}"

    def reverse(self, value: str) -> int:
        """Reverse numeric padding: "005" → 5."""
        try:
            return int(value)
        except ValueError as e:
            msg = f"Cannot parse '{value}' as integer"
            raise MacroResolutionError(
                msg,
                failure_reason=MacroResolutionFailureReason.INVALID_INTEGER_PARSE,
            ) from e


@dataclass
class LowerCaseFormat(FormatSpec):
    """Lowercase transformation :lower."""

    def apply(self, value: str | int) -> str:
        """Convert value to lowercase."""
        return str(value).lower()

    def reverse(self, value: str) -> str:
        """Cannot reliably reverse case - return as-is."""
        return value


@dataclass
class UpperCaseFormat(FormatSpec):
    """Uppercase transformation :upper."""

    def apply(self, value: str | int) -> str:
        """Convert value to uppercase."""
        return str(value).upper()

    def reverse(self, value: str) -> str:
        """Cannot reliably reverse case - return as-is."""
        return value


@dataclass
class SlugFormat(FormatSpec):
    """Slugification format :slug (spaces to hyphens, safe chars only)."""

    def apply(self, value: str | int) -> str:
        """Convert to slug: spaces→hyphens, lowercase, safe chars."""
        s = str(value).lower()
        s = re.sub(r"\s+", "-", s)  # Spaces to hyphens
        s = re.sub(r"[^a-z0-9\-_]", "", s)  # Keep only safe chars
        return s

    def reverse(self, value: str) -> str:
        """Cannot reliably reverse slugification - return as-is."""
        return value


@dataclass
class DateFormat(FormatSpec):
    """Date formatting like :%Y-%m-%d."""

    pattern: str  # e.g., "%Y-%m-%d"

    def apply(self, _value: str | int) -> str:
        """Apply date formatting."""
        # TODO(https://github.com/griptape-ai/griptape-nodes/issues/2717): Implement date formatting
        msg = "DateFormat not yet fully implemented"
        raise MacroResolutionError(
            msg,
            failure_reason=MacroResolutionFailureReason.DATE_FORMAT_NOT_IMPLEMENTED,
        )

    def reverse(self, value: str) -> str:
        """Attempt to parse date string."""
        # TODO(https://github.com/griptape-ai/griptape-nodes/issues/2717): Implement date parsing
        return value


# Module-level registry of known format transformations
FORMAT_REGISTRY: dict[str, FormatSpec] = {
    "lower": LowerCaseFormat(),
    "upper": UpperCaseFormat(),
    "slug": SlugFormat(),
}



"""API client for Nodes API communication."""

from griptape_nodes.api_client.client import Client
from griptape_nodes.api_client.request_client import RequestClient

__all__ = [
    "Client",
    "RequestClient",
]



"""MCP (Model Context Protocol) server management.

Handles MCP server configurations, enabling/disabling servers, and provides
event-based interface for frontend and backend interactions.
"""

import logging
from typing import Any

from griptape_nodes.retained_mode.events.mcp_events import (
    CreateMCPServerRequest,
    CreateMCPServerResultFailure,
    CreateMCPServerResultSuccess,
    DeleteMCPServerRequest,
    DeleteMCPServerResultFailure,
    DeleteMCPServerResultSuccess,
    DisableMCPServerRequest,
    DisableMCPServerResultFailure,
    DisableMCPServerResultSuccess,
    EnableMCPServerRequest,
    EnableMCPServerResultFailure,
    EnableMCPServerResultSuccess,
    GetEnabledMCPServersRequest,
    GetEnabledMCPServersResultFailure,
    GetEnabledMCPServersResultSuccess,
    GetMCPServerRequest,
    GetMCPServerResultFailure,
    GetMCPServerResultSuccess,
    ListMCPServersRequest,
    ListMCPServersResultFailure,
    ListMCPServersResultSuccess,
    UpdateMCPServerRequest,
    UpdateMCPServerResultFailure,
    UpdateMCPServerResultSuccess,
)
from griptape_nodes.retained_mode.managers.config_manager import ConfigManager
from griptape_nodes.retained_mode.managers.event_manager import EventManager
from griptape_nodes.retained_mode.managers.settings import MCPServerConfig

logger = logging.getLogger("griptape_nodes")


class MCPManager:
    """Manager for MCP server configurations and operations."""

    def __init__(self, event_manager: EventManager | None = None, config_manager: ConfigManager | None = None) -> None:
        """Initialize the MCPManager.

        Args:
            event_manager: The EventManager instance to use for event handling.
            config_manager: The ConfigManager instance to use for configuration management.
        """
        self.config_manager = config_manager
        if event_manager is not None:
            # Register event handlers
            event_manager.assign_manager_to_request_type(ListMCPServersRequest, self.on_list_mcp_servers_request)
            event_manager.assign_manager_to_request_type(GetMCPServerRequest, self.on_get_mcp_server_request)
            event_manager.assign_manager_to_request_type(CreateMCPServerRequest, self.on_create_mcp_server_request)
            event_manager.assign_manager_to_request_type(UpdateMCPServerRequest, self.on_update_mcp_server_request)
            event_manager.assign_manager_to_request_type(DeleteMCPServerRequest, self.on_delete_mcp_server_request)
            event_manager.assign_manager_to_request_type(EnableMCPServerRequest, self.on_enable_mcp_server_request)
            event_manager.assign_manager_to_request_type(DisableMCPServerRequest, self.on_disable_mcp_server_request)
            event_manager.assign_manager_to_request_type(
                GetEnabledMCPServersRequest, self.on_get_enabled_mcp_servers_request
            )

    def _get_mcp_servers(self, filter_by: dict[str, Any] | None = None) -> list[MCPServerConfig]:
        """Get the current MCP servers configuration from the config manager.

        Args:
            filter_by: Optional dict of field=value pairs to filter by.
                      Keys should match server config field names, values are the expected values.
        """
        if self.config_manager is None:
            return []

        mcp_config_data = self.config_manager.get_config_value("mcp_servers", default=[])
        if not mcp_config_data:
            return []

        try:
            servers = [MCPServerConfig.model_validate(server) for server in mcp_config_data]
            if filter_by:
                filtered_servers = []
                for server in servers:
                    match = True
                    for field, value in filter_by.items():
                        if getattr(server, field, None) != value:
                            match = False
                            break
                    if match:
                        filtered_servers.append(server)
                return filtered_servers
            return servers  # noqa: TRY300
        except Exception as e:
            logger.error("Failed to parse MCP servers configuration: %s", e)
            return []

    def _save_mcp_servers(self, servers: list[MCPServerConfig]) -> None:
        """Save the MCP servers configuration to the config manager."""
        if self.config_manager is None:
            logger.warning("No config manager available, cannot save MCP configuration")
            return

        try:
            self.config_manager.set_config_value("mcp_servers", [server.model_dump() for server in servers])
        except Exception as e:
            logger.error("Failed to save MCP servers configuration: %s", e)

    def _update_server_fields(self, server_config: MCPServerConfig, request: UpdateMCPServerRequest) -> None:
        """Update server configuration fields from request."""
        # Map request fields to server config attributes
        field_mapping = {
            "new_name": "name",
            "transport": "transport",
            "enabled": "enabled",
            "command": "command",
            "args": "args",
            "env": "env",
            "cwd": "cwd",
            "encoding": "encoding",
            "encoding_error_handler": "encoding_error_handler",
            "url": "url",
            "headers": "headers",
            "timeout": "timeout",
            "sse_read_timeout": "sse_read_timeout",
            "terminate_on_close": "terminate_on_close",
            "description": "description",
            "capabilities": "capabilities",
        }

        # Update fields that are not None
        for request_field, config_field in field_mapping.items():
            value = getattr(request, request_field, None)
            if value is not None:
                setattr(server_config, config_field, value)

    def on_list_mcp_servers_request(
        self, request: ListMCPServersRequest
    ) -> ListMCPServersResultSuccess | ListMCPServersResultFailure:
        """Handle list MCP servers request."""
        try:
            servers = self._get_mcp_servers()
        except Exception as e:
            logger.error("Failed to list MCP servers: %s", e)
            return ListMCPServersResultFailure(result_details=f"Failed to list MCP servers: {e}")

        if request.include_disabled:
            servers_dict = {server.name: server.model_dump() for server in servers}
        else:
            enabled_servers = [server for server in servers if server.enabled]
            servers_dict = {server.name: server.model_dump() for server in enabled_servers}

        # Success path after exception handling
        return ListMCPServersResultSuccess(
            # Pydantic model.model_dump() returns dict[str, Any] which matches MCPServerConfig TypedDict structure
            servers=servers_dict,  # type: ignore[arg-type]
            result_details=f"Successfully listed {len(servers_dict)} MCP servers",
        )

    def on_get_mcp_server_request(
        self, request: GetMCPServerRequest
    ) -> GetMCPServerResultSuccess | GetMCPServerResultFailure:
        """Handle get MCP server request."""
        servers = self._get_mcp_servers(filter_by={"name": request.name})
        server_config = servers[0] if servers else None

        if server_config is None:
            return GetMCPServerResultFailure(result_details=f"Failed to get MCP server '{request.name}' - not found")

        # Success path after exception handling
        return GetMCPServerResultSuccess(
            # Pydantic model.model_dump() returns dict[str, Any] which matches MCPServerConfig TypedDict structure
            server_config=server_config.model_dump(),  # type: ignore[arg-type]
            result_details=f"Successfully retrieved MCP server '{request.name}'",
        )

    def on_create_mcp_server_request(
        self, request: CreateMCPServerRequest
    ) -> CreateMCPServerResultSuccess | CreateMCPServerResultFailure:
        """Handle create MCP server request."""
        try:
            servers = self._get_mcp_servers()
        except Exception as e:
            logger.error("Failed to create MCP server '%s': %s", request.name, e)
            return CreateMCPServerResultFailure(result_details=f"Failed to create MCP server '{request.name}': {e}")

        # Check if server already exists
        for server in servers:
            if server.name == request.name:
                return CreateMCPServerResultFailure(
                    result_details=f"Failed to create MCP server '{request.name}' - already exists"
                )

        # Create new server configuration
        server_config = MCPServerConfig(
            name=request.name,
            enabled=request.enabled,
            transport=request.transport,
            # StdioConnection fields
            command=request.command,
            args=request.args or [],
            env=request.env or {},
            cwd=request.cwd,
            encoding=request.encoding,
            encoding_error_handler=request.encoding_error_handler,
            # HTTP-based connection fields
            url=request.url,
            headers=request.headers,
            timeout=request.timeout,
            sse_read_timeout=request.sse_read_timeout,
            terminate_on_close=request.terminate_on_close,
            # Common fields
            description=request.description,
            capabilities=request.capabilities or [],
        )

        servers.append(server_config)

        try:
            self._save_mcp_servers(servers)
        except Exception as e:
            logger.error("Failed to save MCP server '%s': %s", request.name, e)
            return CreateMCPServerResultFailure(result_details=f"Failed to save MCP server '{request.name}': {e}")

        # Success path after exception handling
        return CreateMCPServerResultSuccess(
            name=request.name, result_details=f"Successfully created MCP server '{request.name}'"
        )

    def on_update_mcp_server_request(
        self, request: UpdateMCPServerRequest
    ) -> UpdateMCPServerResultSuccess | UpdateMCPServerResultFailure:
        """Handle update MCP server request."""
        servers = self._get_mcp_servers()

        # Find the server to update
        server_index = next((i for i, server in enumerate(servers) if server.name == request.name), None)

        if server_index is None:
            return UpdateMCPServerResultFailure(
                result_details=f"Failed to update MCP server '{request.name}' - not found"
            )

        # Create a backup of the original server and update a copy
        original_server = servers[server_index]
        updated_server = original_server.model_copy()
        self._update_server_fields(updated_server, request)

        # Create a copy of the servers list with the updated server
        updated_servers = servers.copy()
        updated_servers[server_index] = updated_server

        try:
            self._save_mcp_servers(updated_servers)
        except Exception as e:
            logger.error("Failed to save MCP server '%s': %s", request.name, e)
            return UpdateMCPServerResultFailure(result_details=f"Failed to save MCP server '{request.name}': {e}")

        # Success path after exception handling
        return UpdateMCPServerResultSuccess(
            name=request.name, result_details=f"Successfully updated MCP server '{request.name}'"
        )

    def on_delete_mcp_server_request(
        self, request: DeleteMCPServerRequest
    ) -> DeleteMCPServerResultSuccess | DeleteMCPServerResultFailure:
        """Handle delete MCP server request."""
        try:
            servers = self._get_mcp_servers()

            # Find and remove server by server_id
            server_found = False
            for i, server in enumerate(servers):
                if server.name == request.name:
                    servers.pop(i)
                    server_found = True
                    break

            self._save_mcp_servers(servers)

        except Exception as e:
            logger.error("Failed to delete MCP server '%s': %s", request.name, e)
            return DeleteMCPServerResultFailure(result_details=f"Failed to delete MCP server '{request.name}': {e}")

        if not server_found:
            return DeleteMCPServerResultFailure(
                result_details=f"Failed to delete MCP server '{request.name}' - not found"
            )

        # Success path after exception handling
        return DeleteMCPServerResultSuccess(
            name=request.name, result_details=f"Successfully deleted MCP server '{request.name}'"
        )

    def on_enable_mcp_server_request(
        self, request: EnableMCPServerRequest
    ) -> EnableMCPServerResultSuccess | EnableMCPServerResultFailure:
        """Handle enable MCP server request."""
        servers = self._get_mcp_servers()
        server_config = None
        for server in servers:
            if server.name == request.name:
                server_config = server
                break

        if server_config is None:
            return EnableMCPServerResultFailure(
                result_details=f"Failed to enable MCP server '{request.name}' - not found"
            )

        server_config.enabled = True

        try:
            self._save_mcp_servers(servers)
        except Exception as e:
            logger.error("Failed to save MCP server '%s': %s", request.name, e)
            return EnableMCPServerResultFailure(result_details=f"Failed to save MCP server '{request.name}': {e}")

        # Success path after exception handling
        return EnableMCPServerResultSuccess(
            name=request.name, result_details=f"Successfully enabled MCP server '{request.name}'"
        )

    def on_disable_mcp_server_request(
        self, request: DisableMCPServerRequest
    ) -> DisableMCPServerResultSuccess | DisableMCPServerResultFailure:
        """Handle disable MCP server request."""
        servers = self._get_mcp_servers()
        server_config = None
        for server in servers:
            if server.name == request.name:
                server_config = server
                break

        if server_config is None:
            return DisableMCPServerResultFailure(
                result_details=f"Failed to disable MCP server '{request.name}' - not found"
            )

        server_config.enabled = False

        try:
            self._save_mcp_servers(servers)
        except Exception as e:
            logger.error("Failed to save MCP server '%s': %s", request.name, e)
            return DisableMCPServerResultFailure(result_details=f"Failed to save MCP server '{request.name}': {e}")

        # Success path after exception handling
        return DisableMCPServerResultSuccess(
            name=request.name, result_details=f"Successfully disabled MCP server '{request.name}'"
        )

    def on_get_enabled_mcp_servers_request(
        self,
        request: GetEnabledMCPServersRequest,  # noqa: ARG002
    ) -> GetEnabledMCPServersResultSuccess | GetEnabledMCPServersResultFailure:
        """Handle get enabled MCP servers request."""
        try:
            enabled_servers = self._get_mcp_servers(filter_by={"enabled": True})
            servers_dict = {server.name: server.model_dump() for server in enabled_servers}

        except Exception as e:
            logger.error("Failed to get enabled MCP servers: %s", e)
            return GetEnabledMCPServersResultFailure(result_details=f"Failed to get enabled MCP servers: {e}")

        # Success path after exception handling
        return GetEnabledMCPServersResultSuccess(
            # Pydantic model.model_dump() returns dict[str, Any] which matches MCPServerConfig TypedDict structure
            servers=servers_dict,  # type: ignore[arg-type]
            result_details=f"Successfully retrieved {len(servers_dict)} enabled MCP servers",
        )



import logging
from typing import NamedTuple

from griptape_nodes.retained_mode.events.base_events import ResultPayload
from griptape_nodes.retained_mode.events.variable_events import (
    CreateVariableRequest,
    CreateVariableResultFailure,
    CreateVariableResultSuccess,
    DeleteVariableRequest,
    DeleteVariableResultFailure,
    DeleteVariableResultSuccess,
    GetVariableDetailsRequest,
    GetVariableDetailsResultFailure,
    GetVariableDetailsResultSuccess,
    GetVariableRequest,
    GetVariableResultFailure,
    GetVariableResultSuccess,
    GetVariableTypeRequest,
    GetVariableTypeResultFailure,
    GetVariableTypeResultSuccess,
    GetVariableValueRequest,
    GetVariableValueResultFailure,
    GetVariableValueResultSuccess,
    HasVariableRequest,
    HasVariableResultFailure,
    HasVariableResultSuccess,
    ListVariablesRequest,
    ListVariablesResultFailure,
    ListVariablesResultSuccess,
    RenameVariableRequest,
    RenameVariableResultFailure,
    RenameVariableResultSuccess,
    SetVariableTypeRequest,
    SetVariableTypeResultFailure,
    SetVariableTypeResultSuccess,
    SetVariableValueRequest,
    SetVariableValueResultFailure,
    SetVariableValueResultSuccess,
    VariableDetails,
)
from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes
from griptape_nodes.retained_mode.managers.event_manager import EventManager
from griptape_nodes.retained_mode.variable_types import FlowVariable, VariableScope

logger = logging.getLogger("griptape_nodes")


class VariableLookupResult(NamedTuple):
    """Result of hierarchical variable lookup."""

    variable: FlowVariable | None
    found_scope: VariableScope | None


class VariablesManager:
    """Manager for variables with scoped access control."""

    def __init__(self, event_manager: EventManager | None = None) -> None:
        # Storage for flow-scoped variables: {flow_name: {variable_name: FlowVariable}}
        self._flow_variables: dict[str, dict[str, FlowVariable]] = {}
        # Storage for global variables: {variable_name: FlowVariable}
        self._global_variables: dict[str, FlowVariable] = {}
        if event_manager is not None:
            event_manager.assign_manager_to_request_type(CreateVariableRequest, self.on_create_variable_request)
            event_manager.assign_manager_to_request_type(GetVariableRequest, self.on_get_variable_request)
            event_manager.assign_manager_to_request_type(GetVariableValueRequest, self.on_get_variable_value_request)
            event_manager.assign_manager_to_request_type(SetVariableValueRequest, self.on_set_variable_value_request)
            event_manager.assign_manager_to_request_type(GetVariableTypeRequest, self.on_get_variable_type_request)
            event_manager.assign_manager_to_request_type(SetVariableTypeRequest, self.on_set_variable_type_request)
            event_manager.assign_manager_to_request_type(DeleteVariableRequest, self.on_delete_variable_request)
            event_manager.assign_manager_to_request_type(RenameVariableRequest, self.on_rename_variable_request)
            event_manager.assign_manager_to_request_type(HasVariableRequest, self.on_has_variable_request)
            event_manager.assign_manager_to_request_type(ListVariablesRequest, self.on_list_variables_request)
            event_manager.assign_manager_to_request_type(
                GetVariableDetailsRequest, self.on_get_variable_details_request
            )

    def on_clear_object_state(self) -> None:
        """Clear all variables."""
        self._flow_variables.clear()
        self._global_variables.clear()

    def _get_starting_flow(self, starting_flow: str | None) -> str:
        """Get the starting flow name, using Context Manager if None."""
        if starting_flow is not None:
            # Validate that the specified flow exists
            flow_manager = GriptapeNodes.FlowManager()
            try:
                flow_manager.get_parent_flow(starting_flow)  # This will raise if flow doesn't exist
            except Exception as e:
                msg = f"Specified starting flow '{starting_flow}' does not exist: {e}"
                raise ValueError(msg) from e
            return starting_flow

        # Get current flow from Context Manager
        context_manager = GriptapeNodes.ContextManager()

        if not context_manager.has_current_flow():
            msg = "No starting flow specified and no current flow in Context Manager"
            raise ValueError(msg)

        return context_manager.get_current_flow().name

    def _get_flow_hierarchy(self, starting_flow: str) -> list[str]:
        """Get the flow hierarchy from starting flow up to root."""
        flow_manager = GriptapeNodes.FlowManager()

        hierarchy = []
        current_flow = starting_flow

        while current_flow:
            hierarchy.append(current_flow)
            try:
                parent = flow_manager.get_parent_flow(current_flow)
                current_flow = parent
            except Exception:
                # No parent flow found, we've reached the root
                break

        return hierarchy

    def _find_variable_in_flow(self, flow_name: str, variable_name: str) -> FlowVariable | None:
        """Find a variable in a specific flow."""
        flow_vars = self._flow_variables.get(flow_name, {})
        return flow_vars.get(variable_name)

    def _find_variable_hierarchical(
        self, starting_flow: str, variable_name: str, lookup_scope: VariableScope
    ) -> VariableLookupResult:
        """Find a variable using hierarchical lookup strategy."""
        match lookup_scope:
            case VariableScope.CURRENT_FLOW_ONLY:
                variable = self._find_variable_in_flow(starting_flow, variable_name)
                found_scope = VariableScope.CURRENT_FLOW_ONLY if variable else None
                return VariableLookupResult(variable=variable, found_scope=found_scope)

            case VariableScope.GLOBAL_ONLY:
                variable = self._global_variables.get(variable_name)
                found_scope = VariableScope.GLOBAL_ONLY if variable else None
                return VariableLookupResult(variable=variable, found_scope=found_scope)

            case VariableScope.HIERARCHICAL:
                # Search through flow hierarchy
                hierarchy = self._get_flow_hierarchy(starting_flow)
                for flow_name in hierarchy:
                    variable = self._find_variable_in_flow(flow_name, variable_name)
                    if variable:
                        found_scope = (
                            VariableScope.CURRENT_FLOW_ONLY
                            if flow_name == starting_flow
                            else VariableScope.HIERARCHICAL
                        )
                        return VariableLookupResult(variable=variable, found_scope=found_scope)

                # Check global variables as fallback
                variable = self._global_variables.get(variable_name)
                found_scope = VariableScope.GLOBAL_ONLY if variable else None
                return VariableLookupResult(variable=variable, found_scope=found_scope)

            case VariableScope.ALL:
                # This is primarily for ListVariables - just search current flow for now
                variable = self._find_variable_in_flow(starting_flow, variable_name)
                found_scope = VariableScope.CURRENT_FLOW_ONLY if variable else None
                return VariableLookupResult(variable=variable, found_scope=found_scope)

            case _:
                msg = f"Attempted to find variable '{variable_name}' from starting flow '{starting_flow}', but encountered an unknown/unexpected variable scope '{lookup_scope.value}'"
                raise ValueError(msg)

    def on_create_variable_request(self, request: CreateVariableRequest) -> ResultPayload:
        """Create a new variable."""
        if request.is_global:
            # Check for name collision in global variables
            if request.name in self._global_variables:
                return CreateVariableResultFailure(
                    result_details=f"Attempted to create a global variable named '{request.name}'. Failed because a variable with that name already exists."
                )

            # Create global variable
            variable = FlowVariable(
                name=request.name,
                owning_flow_name=None,
                type=request.type,
                value=request.value,
            )

            self._global_variables[request.name] = variable
            return CreateVariableResultSuccess(result_details=f"Successfully created global variable '{request.name}'.")

        # Get the target flow
        try:
            target_flow = self._get_starting_flow(request.owning_flow)
        except ValueError as e:
            return CreateVariableResultFailure(
                result_details=f"Attempted to create variable '{request.name}'. Failed to determine target flow: {e}"
            )

        # Initialize flow storage if needed
        if target_flow not in self._flow_variables:
            self._flow_variables[target_flow] = {}

        # Check for name collision in target flow
        if request.name in self._flow_variables[target_flow]:
            return CreateVariableResultFailure(
                result_details=f"Attempted to create a variable named '{request.name}' in flow '{target_flow}'. Failed because a variable with that name already exists."
            )

        # Create flow-scoped variable
        variable = FlowVariable(
            name=request.name,
            owning_flow_name=target_flow,
            type=request.type,
            value=request.value,
        )

        self._flow_variables[target_flow][request.name] = variable
        return CreateVariableResultSuccess(
            result_details=f"Successfully created variable '{request.name}' in flow '{target_flow}'."
        )

    def on_get_variable_request(self, request: GetVariableRequest) -> ResultPayload:
        """Get a full variable by name."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return GetVariableResultFailure(
                result_details=f"Attempted to get variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return GetVariableResultFailure(
                result_details=f"Attempted to get variable '{request.name}'. Failed because no such variable could be found."
            )

        return GetVariableResultSuccess(
            variable=result.variable, result_details=f"Successfully retrieved variable '{request.name}'."
        )

    def on_get_variable_value_request(self, request: GetVariableValueRequest) -> ResultPayload:
        """Get the value of a variable."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return GetVariableValueResultFailure(
                result_details=f"Attempted to get value for variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return GetVariableValueResultFailure(
                result_details=f"Attempted to get value for variable '{request.name}'. Failed because no such variable could be found."
            )

        return GetVariableValueResultSuccess(
            value=result.variable.value, result_details=f"Successfully retrieved value for variable '{request.name}'."
        )

    def on_set_variable_value_request(self, request: SetVariableValueRequest) -> ResultPayload:
        """Set the value of an existing variable."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return SetVariableValueResultFailure(
                result_details=f"Attempted to set value for variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return SetVariableValueResultFailure(
                result_details=f"Attempted to set value for variable '{request.name}'. Failed because no such variable could be found."
            )

        result.variable.value = request.value
        return SetVariableValueResultSuccess(result_details=f"Successfully set value for variable '{request.name}'.")

    def on_get_variable_type_request(self, request: GetVariableTypeRequest) -> ResultPayload:
        """Get the type of a variable."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return GetVariableTypeResultFailure(
                result_details=f"Attempted to get type for variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return GetVariableTypeResultFailure(
                result_details=f"Attempted to get type for variable '{request.name}'. Failed because no such variable could be found."
            )

        return GetVariableTypeResultSuccess(
            type=result.variable.type, result_details=f"Successfully retrieved type for variable '{request.name}'."
        )

    def on_set_variable_type_request(self, request: SetVariableTypeRequest) -> ResultPayload:
        """Set the type of an existing variable."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return SetVariableTypeResultFailure(
                result_details=f"Attempted to set type for variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return SetVariableTypeResultFailure(
                result_details=f"Attempted to set type for variable '{request.name}'. Failed because no such variable could be found."
            )

        result.variable.type = request.type
        return SetVariableTypeResultSuccess(
            result_details=f"Successfully set type for variable '{request.name}' to '{request.type}'."
        )

    def on_delete_variable_request(self, request: DeleteVariableRequest) -> ResultPayload:
        """Delete a variable."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return DeleteVariableResultFailure(
                result_details=f"Attempted to delete variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return DeleteVariableResultFailure(
                result_details=f"Attempted to delete variable '{request.name}'. Failed because no such variable could be found."
            )

        variable = result.variable

        # Remove from appropriate storage based on owning flow
        if variable.owning_flow_name is None:
            # Global variable
            del self._global_variables[variable.name]
        else:
            # Flow-scoped variable
            flow_vars = self._flow_variables.get(variable.owning_flow_name, {})
            if variable.name in flow_vars:
                del flow_vars[variable.name]

        return DeleteVariableResultSuccess(result_details=f"Successfully deleted variable '{request.name}'.")

    def on_rename_variable_request(self, request: RenameVariableRequest) -> ResultPayload:
        """Rename a variable."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return RenameVariableResultFailure(
                result_details=f"Attempted to rename variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return RenameVariableResultFailure(
                result_details=f"Attempted to rename variable '{request.name}'. Failed because no such variable could be found."
            )

        variable = result.variable

        # Check for name collision with new name in the same scope
        new_name_result = self._find_variable_hierarchical(starting_flow, request.new_name, request.lookup_scope)
        if new_name_result.variable and new_name_result.variable.name != variable.name:
            return RenameVariableResultFailure(
                result_details=f"Attempted to rename variable '{request.name}' to '{request.new_name}'. Failed because a variable with that name already exists."
            )

        # Update the variable name and storage key
        old_name = variable.name
        variable.name = request.new_name

        # Update in appropriate storage based on owning flow
        if variable.owning_flow_name is None:
            # Global variable
            del self._global_variables[old_name]
            self._global_variables[request.new_name] = variable
        else:
            # Flow-scoped variable
            flow_vars = self._flow_variables.get(variable.owning_flow_name, {})
            if old_name in flow_vars:
                del flow_vars[old_name]
                flow_vars[request.new_name] = variable

        return RenameVariableResultSuccess(
            result_details=f"Successfully renamed variable '{old_name}' to '{request.new_name}'."
        )

    def on_has_variable_request(self, request: HasVariableRequest) -> ResultPayload:
        """Check if a variable exists."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return HasVariableResultFailure(
                result_details=f"Attempted to check existence of variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)
        exists = result.variable is not None

        return HasVariableResultSuccess(
            exists=exists,
            found_scope=result.found_scope,
            result_details=f"Successfully checked existence of variable '{request.name}': {'exists' if exists else 'not found'}.",
        )

    def _get_variables_by_scope(self, starting_flow: str, lookup_scope: VariableScope) -> list[FlowVariable]:
        """Get variables for the specified scope."""
        match lookup_scope:
            case VariableScope.CURRENT_FLOW_ONLY:
                if starting_flow in self._flow_variables:
                    return list(self._flow_variables[starting_flow].values())
                return []

            case VariableScope.GLOBAL_ONLY:
                return list(self._global_variables.values())

            case VariableScope.HIERARCHICAL:
                return self._get_hierarchical_variables(starting_flow)

            case VariableScope.ALL:
                return self._get_all_variables()

            case _:
                msg = f"Attempted to get variables from starting flow '{starting_flow}', but encountered an unknown/unexpected variable scope '{lookup_scope.value}'"
                raise ValueError(msg)

    def _get_hierarchical_variables(self, starting_flow: str) -> list[FlowVariable]:
        """Get variables using hierarchical lookup with shadowing.

        Variable shadowing behavior:
        - Child flow variables shadow (hide) parent flow variables with same name
        - Flow variables shadow global variables with same name

        Example:
        - Global: user_id = "global_user"
        - Parent flow: user_id = "parent_user"
        - Child flow: user_id = "child_user"
        Result from child flow: only user_id = "child_user" (others are shadowed)
        """
        hierarchy = self._get_flow_hierarchy(starting_flow)
        seen_names = set()
        variables = []

        # Add variables from flows (child to parent to implement shadowing)
        for flow_name in hierarchy:
            flow_vars = self._flow_variables.get(flow_name, {})
            for var in flow_vars.values():
                if var.name not in seen_names:
                    variables.append(var)
                    seen_names.add(var.name)

        # Add global variables (lowest priority, can be shadowed by flow variables)
        variables.extend(var for var in self._global_variables.values() if var.name not in seen_names)

        return variables

    def _get_all_variables(self) -> list[FlowVariable]:
        """Get all variables from all flows for GUI enumeration.

        Note: This returns ALL variables without shadowing - variables with the same name
        from different flows/scopes will all be included in the result.
        """
        variables = []

        # Add all flow variables (no shadowing - include all)
        for flow_vars in self._flow_variables.values():
            variables.extend(flow_vars.values())

        # Add all global variables
        variables.extend(self._global_variables.values())

        return variables

    def on_list_variables_request(self, request: ListVariablesRequest) -> ResultPayload:
        """List all variables in the specified scope."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return ListVariablesResultFailure(
                result_details=f"Attempted to list variables. Failed to determine starting flow: {e}"
            )

        variables = self._get_variables_by_scope(starting_flow, request.lookup_scope)

        # Sort by name for consistent output
        variables = sorted(variables, key=lambda v: v.name)
        return ListVariablesResultSuccess(
            variables=variables, result_details=f"Successfully listed {len(variables)} variables."
        )

    def on_get_variable_details_request(self, request: GetVariableDetailsRequest) -> ResultPayload:
        """Get variable details (metadata only, no heavy values)."""
        try:
            starting_flow = self._get_starting_flow(request.starting_flow)
        except ValueError as e:
            return GetVariableDetailsResultFailure(
                result_details=f"Attempted to get details for variable '{request.name}'. Failed to determine starting flow: {e}"
            )

        result = self._find_variable_hierarchical(starting_flow, request.name, request.lookup_scope)

        if not result.variable:
            return GetVariableDetailsResultFailure(
                result_details=f"Attempted to get details for variable '{request.name}'. Failed because no such variable could be found."
            )

        variable = result.variable
        details = VariableDetails(name=variable.name, owning_flow_name=variable.owning_flow_name, type=variable.type)
        return GetVariableDetailsResultSuccess(
            details=details, result_details=f"Successfully retrieved details for variable '{request.name}'."
        )

    def _find_variable_by_name(self, name: str) -> FlowVariable | None:
        """Find a variable by name in current flow context (legacy compatibility)."""
        try:
            starting_flow = self._get_starting_flow(None)
        except ValueError:
            return None

        result = self._find_variable_hierarchical(starting_flow, name, VariableScope.HIERARCHICAL)
        return result.variable



import logging

from griptape_nodes.exe_types.core_types import Parameter, ParameterMode
from griptape_nodes.exe_types.node_types import BaseNode

logger = logging.getLogger(__name__)


class ProgressBarComponent:
    def __init__(self, node: BaseNode):
        self._node = node
        self._total_steps = 0
        self._current_step = 0

    def add_property_parameters(self) -> None:
        self._node.add_parameter(
            Parameter(
                name="progress",
                output_type="float",
                allowed_modes={ParameterMode.PROPERTY},
                tooltip="Progress bar showing completion (0.0 to 1.0)",
                ui_options={"progress_bar": True},
                settable=False,
            )
        )

    def initialize(self, total_steps: int) -> None:
        """Initialize the progress bar with a total number of steps."""
        self._total_steps = total_steps
        self._current_step = 0
        self._update_progress()

    def increment(self, steps: int = 1) -> None:
        """Increment the progress by the specified number of steps."""
        self._current_step += steps
        if self._current_step > self._total_steps:
            logger.warning(
                "Current step %i exceeds total steps %i. Progress will not exceed 100%.",
                self._current_step,
                self._total_steps,
            )
        self._update_progress()

    def reset(self) -> None:
        """Reset the progress bar to 0."""
        self._current_step = 0
        self._total_steps = 0
        self._update_progress()

    def _update_progress(self) -> None:
        """Update the progress parameter based on current step and total steps."""
        if self._total_steps <= 0:
            progress_value = 0.0
        else:
            progress_value = min(1.0, self._current_step / self._total_steps)

        self._node.publish_update_to_parameter("progress", progress_value)



from __future__ import annotations

import json
import logging
from datetime import datetime  # noqa: TC003 (can't put into type checking block as Pydantic model relies on it)
from pathlib import Path
from typing import Any, ClassVar, NamedTuple

from pydantic import BaseModel, Field, field_serializer, field_validator

from griptape_nodes.node_library.library_registry import (
    LibraryNameAndVersion,  # noqa: TC001 (putting this into type checking causes it to not be defined)
)
from griptape_nodes.utils.metaclasses import SingletonMeta

logger = logging.getLogger("griptape_nodes")


class LibraryNameAndNodeType(NamedTuple):
    library_name: str
    node_type: str


# Type aliases for clarity
type NodeName = str
type ParameterName = str
type ParameterAttribute = str
type ParameterMinimalDict = dict[ParameterAttribute, Any]
type NodeParametersMapping = dict[NodeName, dict[ParameterName, ParameterMinimalDict]]


class WorkflowShape(BaseModel):
    """This structure reflects the input and output shapes extracted from StartNodes and EndNodes inside of the workflow.

    A workflow may have multiple StartNodes and multiple EndNodes, each contributing their parameters
    to the overall workflow shape.

    Structure is:
    - inputs: {start_node_name: {param_name: param_minimal_dict}}
    - outputs: {end_node_name: {param_name: param_minimal_dict}}
    """

    inputs: NodeParametersMapping = Field(default_factory=dict)
    outputs: NodeParametersMapping = Field(default_factory=dict)


class WorkflowMetadata(BaseModel):
    LATEST_SCHEMA_VERSION: ClassVar[str] = "0.12.0"

    name: str
    schema_version: str
    engine_version_created_with: str
    node_libraries_referenced: list[LibraryNameAndVersion]
    node_types_used: set[LibraryNameAndNodeType] = Field(default_factory=set)
    workflows_referenced: list[str] | None = None
    description: str | None = None
    image: str | None = None
    is_griptape_provided: bool | None = False
    is_template: bool | None = False
    creation_date: datetime | None = Field(default=None)
    last_modified_date: datetime | None = Field(default=None)
    branched_from: str | None = Field(default=None)
    workflow_shape: WorkflowShape | None = Field(default=None)

    @field_serializer("node_types_used")
    def serialize_node_types_used(self, node_types_used: set[LibraryNameAndNodeType]) -> list[list[str]]:
        """Serialize node_types_used as list of tuples for TOML compatibility.

        Sets and NamedTuples are not directly supported by TOML, so we convert the set
        to a list of lists (each inner list represents [library_name, node_type]).
        """
        return [[nt.library_name, nt.node_type] for nt in sorted(node_types_used)]

    @field_validator("node_types_used", mode="before")
    @classmethod
    def validate_node_types_used(cls, value: Any) -> set[LibraryNameAndNodeType]:
        """Deserialize node_types_used from list of lists during TOML loading.

        When loading workflow metadata from TOML files, the node_types_used field
        is stored as a list of [library_name, node_type] pairs that needs to be
        converted back to a set of LibraryNameAndNodeType objects. This validator
        handles the expected input formats:
        - List of lists (from TOML deserialization)
        - Set of LibraryNameAndNodeType (from direct Python construction)
        - Empty list (for workflows with no nodes)
        """
        if isinstance(value, set):
            return value
        if isinstance(value, list):
            return {LibraryNameAndNodeType(library_name=item[0], node_type=item[1]) for item in value}
        msg = f"Expected list or set for node_types_used, got {type(value)}"
        raise ValueError(msg)

    @field_serializer("workflow_shape")
    def serialize_workflow_shape(self, workflow_shape: WorkflowShape | None) -> str | None:
        """Serialize WorkflowShape as JSON string to avoid TOML serialization issues.

        The WorkflowShape contains deeply nested dictionaries with None values that are
        meaningful data (e.g., default_value: None). TOML's nested table format creates
        unreadable output and tomlkit fails on None values in nested structures.
        JSON preserves None as null and keeps the data compact and readable.
        """
        if workflow_shape is None:
            return None
        # Use json.dumps to preserve None values as null, which TOML can handle
        return json.dumps(workflow_shape.model_dump(), separators=(",", ":"))

    @field_validator("workflow_shape", mode="before")
    @classmethod
    def validate_workflow_shape(cls, value: Any) -> WorkflowShape | None:
        """Deserialize WorkflowShape from JSON string during TOML loading.

        When loading workflow metadata from TOML files, the workflow_shape field
        is stored as a JSON string that needs to be converted back to a WorkflowShape
        object. This validator handles the expected input formats:
        - JSON strings (from TOML deserialization)
        - WorkflowShape objects (from direct Python construction)
        - None values (workflows without Start/End nodes)

        If JSON deserialization fails, logs a warning and returns None for graceful
        degradation, consistent with other metadata parsing failures in this codebase.
        """
        if value is None:
            return None
        if isinstance(value, WorkflowShape):
            return value
        if isinstance(value, str):
            try:
                data = json.loads(value)
                return WorkflowShape(**data)
            except (json.JSONDecodeError, TypeError, ValueError) as e:
                logger.error("Failed to deserialize workflow_shape from JSON: %s", e)
                return None
        # Unexpected type - let Pydantic's normal validation handle it
        return value


class WorkflowRegistry(metaclass=SingletonMeta):
    class _RegistryKey:
        """Private class for workflow construction."""

    _workflows: ClassVar[dict[str, Workflow]] = {}
    _registry_key: _RegistryKey = _RegistryKey()

    # Create a new workflow with everything we'd need
    @classmethod
    def generate_new_workflow(cls, file_path: str, metadata: WorkflowMetadata) -> Workflow:
        instance = cls()
        if metadata.name in instance._workflows:
            msg = f"Workflow with name '{metadata.name}' already registered."
            raise KeyError(msg)
        workflow = Workflow(registry_key=instance._registry_key, file_path=file_path, metadata=metadata)
        instance._workflows[metadata.name] = workflow
        return workflow

    @classmethod
    def get_workflow_by_name(cls, name: str) -> Workflow:
        instance = cls()
        if name not in instance._workflows:
            msg = f"Failed to get Workflow. Workflow with name '{name}' has not been registered."
            raise KeyError(msg)
        return instance._workflows[name]

    @classmethod
    def has_workflow_with_name(cls, name: str) -> bool:
        instance = cls()
        return name in instance._workflows

    @classmethod
    def list_workflows(cls) -> dict[str, dict]:
        instance = cls()
        return {key: instance._workflows[key].get_workflow_metadata() for key in instance._workflows}

    @classmethod
    def get_complete_file_path(cls, relative_file_path: str) -> str:
        from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes

        # If the path is already absolute, return it as-is
        if Path(relative_file_path).is_absolute():
            return relative_file_path

        # Otherwise, resolve it relative to the workspace
        config_mgr = GriptapeNodes.ConfigManager()
        workspace_path = config_mgr.workspace_path
        complete_file_path = workspace_path / relative_file_path
        return str(complete_file_path)

    @classmethod
    def delete_workflow_by_name(cls, name: str) -> Workflow:
        instance = cls()
        if name not in instance._workflows:
            msg = f"Failed to delete Workflow. Workflow with name '{name}' has not been registered."
            raise KeyError(msg)
        return instance._workflows.pop(name)

    @classmethod
    def get_branches_of_workflow(cls, workflow_name: str) -> list[str]:
        """Get all workflows that are branches of the specified workflow."""
        instance = cls()
        branches = []
        for name, workflow in instance._workflows.items():
            if workflow.metadata.branched_from == workflow_name:
                branches.append(name)
        return branches


class Workflow:
    """A workflow card to be ran."""

    metadata: WorkflowMetadata
    file_path: str

    def __init__(self, registry_key: WorkflowRegistry._RegistryKey, metadata: WorkflowMetadata, file_path: str) -> None:
        if not isinstance(registry_key, WorkflowRegistry._RegistryKey):
            msg = "Workflows can only be created through WorkflowRegistry"
            raise TypeError(msg)

        self.metadata = metadata
        self.file_path = file_path

        # Get the absolute file path.
        complete_path = WorkflowRegistry.get_complete_file_path(relative_file_path=file_path)
        if not Path(complete_path).is_file():
            msg = f"File path '{complete_path}' does not exist."
            raise ValueError(msg)

    @property
    def is_synced(self) -> bool:
        """Check if this workflow is in the synced workflows directory."""
        from griptape_nodes.retained_mode.griptape_nodes import GriptapeNodes

        config_mgr = GriptapeNodes.ConfigManager()
        synced_directory = config_mgr.get_config_value("synced_workflows_directory")

        # Get the full path to the synced workflows directory
        synced_path = config_mgr.get_full_path(synced_directory)

        # Get the complete file path for this workflow
        complete_file_path = WorkflowRegistry.get_complete_file_path(self.file_path)

        # Check if the workflow file is within the synced directory
        return Path(complete_file_path).is_relative_to(synced_path)

    def get_workflow_metadata(self) -> dict:
        # Convert from the Pydantic schema.
        ret_val = {**self.metadata.model_dump()}

        # The schema doesn't have the file path in it, because it is baked into the file itself.
        # Customers of this function need that, so let's stuff it in.
        ret_val["file_path"] = self.file_path
        ret_val["is_synced"] = self.is_synced
        return ret_val



"""Package for griptape nodes storage drivers.

Storage drivers are responsible for managing the storage and retrieval of files.
"""

from .storage_backend import StorageBackend

__all__ = ["StorageBackend"]
